# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Annotated, ClassVar, Literal, Sequence

from str_to_obj import annotation_t
from str_to_obj.type.hint import annotated_hint_t

from conf_ini_g.catalog.specification.annotation.choices import choices_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class callable_t(annotation_t):
    ACCEPTED_TYPES = (str,)
    DUMMY_ENTRY: ClassVar[str] = "Dummy entry"
    SEPARATOR: ClassVar[str] = ":"
    kind: Literal["class", "function"]
    choices: choices_t | None = None
    allow_external: bool = False

    @classmethod
    def NewAnnotatedType(
        cls,
        /,
        *,
        kind: Literal["class", "function"] = "function",
        catalog: Sequence[str] | bool | None = None,
        allow_external: bool = False,
    ) -> annotated_hint_t:
        """"""
        if catalog is None:
            annotation = cls(kind=kind, allow_external=allow_external)
        elif isinstance(catalog, bool):
            if catalog:
                annotation = cls(
                    kind=kind,
                    choices=choices_t((cls.DUMMY_ENTRY,)),
                    allow_external=allow_external,
                )
            else:
                annotation = cls(kind=kind, allow_external=allow_external)
        else:
            annotation = cls(
                kind=kind, choices=choices_t(catalog), allow_external=allow_external
            )

        return Annotated[str, annotation]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence

from babelwidget.main import backend_t, scroll_container_h
from babelwidget.main import tabs_h as tabs_wgt_h

from conf_ini_g.interface.window.component.section.main import (
    controlled_section_t,
    section_t,
)
from conf_ini_g.phase.specification.config.main import config_t as config_spec_t
from conf_ini_g.phase.specification.section.main import section_t as section_spec_t

any_section_h = section_t | controlled_section_t


def SectionsAndCategories(
    specification: config_spec_t,
    category_selector: scroll_container_h | tabs_wgt_h | None,
    backend: backend_t,
    /,
) -> tuple[
    dict[str, any_section_h],
    Sequence[tuple[controlled_section_t, section_spec_t]],
    scroll_container_h | tabs_wgt_h,
]:
    """"""
    sections = {}
    controlled_sections = []
    categories = {}

    # --- Sections and categories
    for section_spec in specification:
        if (controller := section_spec.controller) is None:
            section = section_t.NewForSpecification(section_spec, backend)
        else:
            section = controlled_section_t.NewForSpecification(
                section_spec,
                controller,
                backend,
            )
            if section is not None:
                controlled_sections.append((section, section_spec))
        if section is None:
            continue

        sections[section_spec.name] = section

        if (category := section_spec.category) not in categories:
            contents = backend.base_t()
            scroll_area = backend.scroll_container_t.NewForWidget(contents)
            layout = backend.vbox_lyt_t()
            contents.setLayout(layout)
            categories[category] = (layout, scroll_area)

        layout = categories[category][0]
        layout.addWidget(section.library_wgt)

    # --- Section dispatch into categories
    missing_category_selector = category_selector is None
    if categories.__len__() > 1:
        if missing_category_selector:
            category_selector = backend.tabs_t()
        for category, (_, scroll_area) in categories.items():
            category_selector.addTab(scroll_area, category)
    elif missing_category_selector:
        category = tuple(categories.keys())[0]
        category_selector = categories[category][1]

    return sections, controlled_sections, category_selector



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t

any_path_h = str | path_t


def ValidateInputPath(
    path: path_t, /, *, should_raise_on_error: bool = False
) -> list[str]:
    """"""
    if not path.exists():
        issue = f"{path}: Non-existing path."
        if should_raise_on_error:
            raise FileNotFoundError(issue)
        else:
            return [issue]

    if path.is_dir():
        issue = f"{path}: Path is a folder; Expected=Path to an existing file."
        if should_raise_on_error:
            raise IsADirectoryError(issue)
        else:
            return [issue]

    return []


def ValidateOutputPath(
    path: path_t,
    /,
    *,
    should_overwrite: bool = False,
    should_raise_on_error: bool = False,
) -> list[str]:
    """"""
    if path.exists() and not should_overwrite:
        issue = f"{path}: Path exists and overwriting is not allowed."
        if should_raise_on_error:
            raise PermissionError(issue)
        else:
            return [issue]

    if path.is_dir():
        if should_overwrite:
            document_qualification = ""
            document_property = ", existing or not"
        else:
            document_qualification = "non-existing "
            document_property = ""
        issue = (
            f"{path}: Path is a folder; "
            f"Expected=Path to a {document_qualification}file{document_property}."
        )
        if should_raise_on_error:
            raise IsADirectoryError(issue)
        else:
            return [issue]

    return []



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from configparser import ConfigParser as config_parser_t

from openpyxl import Workbook as workbook_t

from conf_ini_g.extension.path import any_path_h, path_t
from conf_ini_g.light.ini import NewConfigFromPath as NewConfigFromINIPath
from conf_ini_g.light.xlsx import NewConfigFromPath as NewConfigFromXLSXPath


def INItoXLSX(
    ini_path: any_path_h, xlsx_path: any_path_h, /, *, should_overwrite: bool = False
) -> None:
    """"""
    ini_path, xlsx_path = _AsPaths(ini_path, xlsx_path)
    _CheckOverwriting(xlsx_path, should_overwrite)

    config = NewConfigFromINIPath(ini_path)

    workbook = workbook_t()
    # Remove the (a priori unique) default worksheet
    for worksheet in workbook.sheetnames:
        # The "remove" method is for worksheets, not names. Use this instead:
        del workbook[worksheet]
    for sct_name, content in config.items():
        worksheet = workbook.create_sheet(title=sct_name)
        for prm_name, value in content.items():
            worksheet.append((prm_name, str(value)))

    workbook.save(xlsx_path)


def XLSXtoINI(
    xlsx_path: any_path_h, ini_path: any_path_h, /, *, should_overwrite: bool = False
) -> None:
    """"""
    ini_path, xlsx_path = _AsPaths(ini_path, xlsx_path)
    _CheckOverwriting(ini_path, should_overwrite)

    config_as_dict = NewConfigFromXLSXPath(xlsx_path)
    config_as_dict = {
        _sct: {_prm: str(_vle) for _prm, _vle in _ctn.items()}
        for _sct, _ctn in config_as_dict.items()
    }

    config = config_parser_t()
    config.optionxform = lambda option: option  # To avoid automatic lowercasing
    config.read_dict(config_as_dict)
    with open(ini_path, "w") as accessor:
        config.write(accessor)


def _AsPaths(ini_path: any_path_h, xlsx_path: any_path_h, /) -> tuple[path_t, path_t]:
    """"""
    if isinstance(ini_path, str):
        ini_path = path_t(ini_path)
    if isinstance(xlsx_path, str):
        xlsx_path = path_t(xlsx_path)

    return ini_path, xlsx_path


def _CheckOverwriting(path: path_t, should_overwrite: bool, /) -> None:
    """"""
    if path.exists() and (path.is_dir() or (path.is_file() and not should_overwrite)):
        raise ValueError(f"{path}: Path already exists.")



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from types import NoneType, UnionType

from babelwidget.main import base_h as library_wgt_h
from str_to_obj import annotation_t
from str_to_obj.type.hint_tree import hint_tree_t

from conf_ini_g.catalog.interface.window.parameter.boolean import boolean_wgt_t
from conf_ini_g.catalog.interface.window.parameter.callable import callable_wgt_t
from conf_ini_g.catalog.interface.window.parameter.choices import choices_wgt_t
from conf_ini_g.catalog.interface.window.parameter.collection import collection_wgt_t
from conf_ini_g.catalog.interface.window.parameter.none import none_wgt_t
from conf_ini_g.catalog.interface.window.parameter.path import path_wgt_t
from conf_ini_g.catalog.interface.window.parameter.text_line import text_line_t
from conf_ini_g.catalog.specification.annotation.callable import callable_t
from conf_ini_g.catalog.specification.annotation.choices import choices_t
from conf_ini_g.phase.specification.parameter.type import type_t

# Widgets can be mapped from types or annotations. Since annotations are more specific
# than types, they must be placed first to ensure they get a chance to be selected.
_TYPE_WIDGET_TRANSLATOR: dict[type, type[library_wgt_h]] = {
    # Annotations
    callable_t: callable_wgt_t,
    choices_t: choices_wgt_t,
    # Types
    NoneType: none_wgt_t,
    UnionType: choices_wgt_t,
    bool: boolean_wgt_t,
    float: text_line_t,
    int: text_line_t,
    list: collection_wgt_t,
    path_t: path_wgt_t,
    set: collection_wgt_t,
    str: text_line_t,
    tuple: collection_wgt_t,
}


def WidgetTypeForType(stripe: type_t | hint_tree_t, /) -> type[library_wgt_h]:
    """"""
    base_hint = stripe.type
    nnts = stripe.all_annotations

    for registered_type, widget_type in _TYPE_WIDGET_TRANSLATOR.items():
        if (base_hint is registered_type) or any(
            issubclass(type(_nnt), registered_type) for _nnt in nnts
        ):
            return widget_type

    return text_line_t


def RegisterNewTranslation(
    new_type: type | annotation_t, widget_type: type[library_wgt_h], /
) -> None:
    """"""
    if new_type in _TYPE_WIDGET_TRANSLATOR:
        # Raising an exception is adapted here since it is a developer-oriented function
        raise ValueError(
            f'{new_type.__name__}: Type already registered with "{_TYPE_WIDGET_TRANSLATOR[new_type]}" '
            f"in type-to-widget translations."
        )

    _TYPE_WIDGET_TRANSLATOR[new_type] = widget_type



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any

from conf_ini_g.interface.storage.parameter import INI_COMMENT_MARKER
from conf_ini_g.interface.storage.section import INI_UNIT_SECTION
from conf_ini_g.phase.specification.config.main import config_t as specification_t
from conf_ini_g.phase.typed.parameter.invalid import INVALID_VALUE
from conf_ini_g.phase.typed.parameter.main import TypedValue
from conf_ini_g.phase.typed.parameter.unit import ConvertedValue
from conf_ini_g.phase.typed.section.main import UpdateSpecification
from conf_ini_g.phase.typed.section.units import InitialUnits, UpdateUnits
from conf_ini_g.phase.untyped.config import ini_config_h
from conf_ini_g.phase.untyped.parameter import ValueUnitAndComment

# With interpreted values, and possibly units.
typed_config_h = dict[str, dict[str, Any]]


def TypedFromUntyped(
    ini: ini_config_h,
    cmd_line: ini_config_h | None,
    specification: specification_t,
    /,
) -> tuple[
    typed_config_h,
    dict[tuple[str, str], tuple[str, float]],
    dict[tuple[str, str], str],
    list[str],
]:
    """
    Units are not consumed, but returned.
    """
    output = {}
    output_units = {}
    output_comments = {}

    if cmd_line is None:
        configs = (ini,)
    else:
        configs = (ini, cmd_line)
    issues = []
    units, spec_has_units = InitialUnits(specification)
    for config in configs:
        for sct_name, parameters in config.items():
            if sct_name not in output:
                output[sct_name] = {}

            if (
                section_is_units := (sct_name == INI_UNIT_SECTION)
            ) and not spec_has_units:
                specification.AddUnitSection()
            section_spec = specification[sct_name]

            for prm_name, value in parameters.items():
                if section_is_units:
                    value, local_issues = UpdateUnits(units, prm_name, value)
                    if local_issues.__len__() > 0:
                        issues.extend(local_issues)
                        continue
                if prm_name not in section_spec:
                    value, local_issues = UpdateSpecification(
                        specification, section_spec, sct_name, prm_name, value
                    )
                    if local_issues.__len__() > 0:
                        issues.extend(local_issues)
                        continue
                output[sct_name][prm_name] = value

    for section in specification:
        if section.name not in output:
            output[section.name] = {}

        if section.controller is None:
            # TODO: Iteration over section will be replaced with something else when the class section will be modified.
            parameters = section
        else:
            controlling_value = output[section.controller.section][
                section.controller.parameter
            ]
            parameters = section.ActiveParameters(controlling_value)

        output_section = output[section.name]
        for parameter in parameters:
            if parameter.optional and (parameter.name not in output_section):
                output[section.name][parameter.name] = parameter.default

    for sct_name, parameters in output.items():
        if sct_name == INI_UNIT_SECTION:
            continue

        section_spec = specification[sct_name]
        for prm_name, value in parameters.items():
            if isinstance(value, str):
                value, unit, comment = ValueUnitAndComment(value, INI_COMMENT_MARKER)
                if not ((unit is None) or (unit in units)):
                    issues.append(f"/{sct_name}.{prm_name}/ Unknown unit: {unit}.")
                    continue
                parameter_spec = section_spec[prm_name]
                value, current_issues = TypedValue(value, parameter_spec.type)
                if current_issues.__len__() > 0:
                    issues.extend(
                        f"/{sct_name}.{prm_name}/ {_iss}" for _iss in current_issues
                    )
                    continue
                output[sct_name][prm_name] = value
                if unit is not None:
                    output_units[(sct_name, prm_name)] = (unit, units[unit])
                if comment is not None:
                    output_comments[(sct_name, prm_name)] = comment

    return output, output_units, output_comments, issues


def WithConsumedUnits(
    config: typed_config_h, units: dict[tuple[str, str], tuple[str, float]], /
) -> tuple[typed_config_h, list[str]]:
    """"""
    output = {_key: dict(_vle) for _key, _vle in config.items()}
    issues = []

    for sct_name, parameters in output.items():
        for prm_name, value in parameters.items():
            unit = units.get((sct_name, prm_name), None)
            if unit is not None:
                converted, unconverted = ConvertedValue(
                    output[sct_name][prm_name], unit[1]
                )
                if unconverted.__len__() > 0:
                    converted = INVALID_VALUE
                    unconverted = ", ".join(unconverted)
                    issues.append(
                        f"{unconverted}: Value(s) do(es) not support unit conversion."
                    )

                output[sct_name][prm_name] = converted

    return output, issues



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Annotated, Callable

from babelwidget.main import backend_t, hbox_lyt_h

from conf_ini_g.catalog.interface.window.parameter.boolean import boolean_wgt_t
from conf_ini_g.catalog.specification.annotation.boolean import boolean_t
from conf_ini_g.phase.specification.parameter.type import type_t


def AdvancedModeLayout(
    advanced_mode: bool, backend: backend_t, ToggleAdvancedMode: Callable, /
) -> tuple[hbox_lyt_h, boolean_wgt_t]:
    """"""
    layout = backend.hbox_lyt_t()

    annotated_type = Annotated[bool, boolean_t(mode=boolean_t.MODE.on_off)]
    value_type = type_t.NewFromTypeHint(annotated_type)
    boolean = boolean_wgt_t.NewForSpecification(
        value_type,
        backend,
    )
    boolean.Assign(advanced_mode, None)
    boolean.true_btn.SetFunction(ToggleAdvancedMode)

    layout.addWidget(backend.label_t("<i>Advanced Mode</i>"))
    layout.addWidget(boolean.library_wgt)

    return layout, boolean



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import textwrap as text
from typing import Any

from str_to_obj import CastValue
from str_to_obj.interface.console import NameValueTypeAsRichStr, TypeAsRichStr
from str_to_obj.type.hint import any_hint_h

from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.phase.specification.base import base_t
from conf_ini_g.phase.specification.parameter.type import ANY_TYPE, type_t
from conf_ini_g.phase.specification.parameter.value import (
    MISSING_REQUIRED_VALUE,
    missing_required_value_t,
)


@dtcl.dataclass(repr=False, eq=False)
class parameter_t(base_t):
    """
    type:
        At instantiation time:
            - any_hint_h, if passed
            - type_t (ANY_TYPE), if not.
        After __post_init__: type_t
    """

    type: any_hint_h | type_t = ANY_TYPE
    default: Any = MISSING_REQUIRED_VALUE

    def __post_init__(self) -> None:
        """"""
        if not isinstance(self.type, type_t):
            self.type = type_t.NewFromTypeHint(self.type)

    def Issues(self) -> list[str]:
        """"""
        output = super().Issues()

        if self.optional:
            output.extend(f"/{self.name}/ {_iss}" for _iss in self.type.Issues())
            if self.default is MISSING_REQUIRED_VALUE:
                output.append(
                    f"{self.name}: Default value of optional parameter cannot be of type "
                    f'"{missing_required_value_t.__name__}"'
                )
            else:
                converted, issues = CastValue(self.default, self.type)
                if issues.__len__() > 0:
                    output.append(
                        f"{self.default}: Invalid default value of parameter "
                        f'"{self.name}": {", ".join(issues)}'
                    )
                else:
                    self.default = converted
        else:
            if not self.basic:
                output.append(f"{self.name}: Parameter is not basic but not optional")
            if self.default is not MISSING_REQUIRED_VALUE:
                output.append(
                    f"{self.name}: Default value of mandatory parameter must be of type "
                    f'"{missing_required_value_t.__name__}"'
                )

        output.extend(self.type.Issues())

        return output

    @property
    def optional(self) -> bool:
        """"""
        return self.default is not MISSING_REQUIRED_VALUE

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            *text.indent(super().__rich__(), "    ").splitlines(),
            f"    [blue]Template[/]@=@{self.type.__rich__()}",
            f"    {NameValueTypeAsRichStr('Default', self.default, separator='@=@')}",
            f"    [blue]Optional[/]@=@{self.optional}",
        ]

        output = AlignedOnSeparator(output, "@=@", " = ")

        return "\n".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import inspect as nspt
import typing as h
from importlib import util as mprt
from pathlib import Path as pl_path_t

from babelwidget.main import backend_t
from babelwidget.main import base_h as library_wgt_h

from conf_ini_g.catalog.interface.window.parameter.choices import choices_wgt_t
from conf_ini_g.catalog.interface.window.parameter.path import path_wgt_t
from conf_ini_g.catalog.specification.annotation.callable import callable_t
from conf_ini_g.catalog.specification.annotation.path import path_t
from conf_ini_g.catalog.specification.annotation.path import path_t as path_annotation_t
from conf_ini_g.interface.window.component.parameter.value import value_wgt_a
from conf_ini_g.phase.specification.parameter.type import type_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class callable_wgt_t(value_wgt_a):
    library_wgt: library_wgt_h
    backend: backend_t
    kind: h.Literal["class", "function"]
    path: path_wgt_t | None = dtcl.field(init=False, default=None)
    choices: choices_wgt_t = dtcl.field(init=False, default=None)

    @classmethod
    def NewForSpecification(
        cls,
        value_type: type_t | callable_t,
        backend: backend_t,
        /,
    ) -> callable_wgt_t:
        """"""
        # TODO: Maybe, re-implement NewForSpecification methods so that they build the widget,
        #     and then assign the value with the Assign method.
        if isinstance(value_type, type_t):
            annotation = value_type.FirstAnnotationWithType(callable_t)
            if annotation is None:
                raise ValueError("No catalog provided.")
        else:
            annotation = value_type
        kind = annotation.kind
        choices = annotation.choices
        allow_external = annotation.allow_external

        choices_wgt = choices_wgt_t.NewForSpecification(
            choices,
            backend,
        )

        if allow_external:
            output = cls(
                library_wgt=backend.base_t(),
                backend=backend,
                kind=kind,
            )
        else:
            output = cls(
                library_wgt=choices_wgt,
                backend=backend,
                kind=kind,
            )
        output.choices = choices_wgt

        if allow_external:
            path_wgt = path_wgt_t.NewForSpecification(
                path_annotation_t(
                    target_type=path_annotation_t.TARGET_TYPE.document, is_input=True
                ),
                backend,
                PostAssignmentFunction=output._UpdateChoices,
            )
            output.path = path_wgt

            layout = backend.hbox_lyt_t()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(path_wgt.library_wgt)
            layout.addWidget(choices_wgt.library_wgt)
            output.library_wgt.setLayout(layout)

        return output

    def _UpdateChoices(self, path: pl_path_t, /) -> None:
        """"""
        self.choices.library_wgt.clear()

        spec = mprt.spec_from_file_location(path.stem, path)
        module = spec.loader.load_module(spec.name)
        if self.kind == "class":
            condition = nspt.isclass
        else:
            condition = nspt.isfunction

        for name, _ in nspt.getmembers(module, condition):
            if not name.startswith("_"):
                self.choices.library_wgt.addItem(name)

    def SetFunction(self, function: h.Callable, /) -> None:
        """
        self.activated.connect(function): Responds only to GUI interaction, not
        programmatic change.
        """
        self.choices.currentIndexChanged.connect(function)

    def Assign(self, value: str, _: type_t | path_t | None, /) -> None:
        """"""
        # TODO: Check, and explain, why Assign needs value_type, potentially.
        if callable_t.SEPARATOR in value:
            path, choice = value.split(callable_t.SEPARATOR)
            self.path.Assign(pl_path_t(path), None)
        else:
            choice = value
        self.choices.Assign(choice, None)

    def Text(self) -> str:
        """"""
        if self.path is None:
            return self.choices.Text()

        path = self.path.Text()
        if path.__len__() > 0:
            return f"{path}{callable_t.SEPARATOR}{self.choices.Text()}"

        return self.choices.Text()



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Any

from rich.text import Text as text_t
from str_to_obj import CastValue, ObjectFromStr, annotation_t
from str_to_obj.type.hint_tree import hint_tree_t

from conf_ini_g.phase.typed.parameter.invalid import INVALID_VALUE


@dtcl.dataclass(slots=True, repr=False, eq=False)
class type_t(hint_tree_t):
    def Issues(self) -> list[str]:
        """"""
        output = []

        for annotation in self.all_annotations:
            output.extend(
                f"/{type(self).__name__}.{id(self)}/ {_iss}"
                for _iss in annotation.Issues()
            )

        return output

    def FirstAnnotationWithType(
        self, stripe: type[annotation_t], /
    ) -> annotation_t | None:
        """"""
        for annotation in self.all_annotations:
            if isinstance(annotation, stripe):
                return annotation

        return None

    def ValueIsCompliant(self, value: Any, /) -> list[str]:
        """"""
        return CastValue(value, self, only_check_validity=True)

    def InterpretedValueOf(self, value: Any, /) -> tuple[Any, list[str]]:
        """"""
        if isinstance(value, str):
            typed_value, issues = ObjectFromStr(value, expected_type=self)
        else:
            typed_value, issues = CastValue(value, self)

        if issues.__len__() > 0:
            return INVALID_VALUE, issues

        return typed_value, []

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        output = [f"[white]{self.template_as_str}[/]"]

        for annotation in self.all_annotations:
            output.append(type(annotation).__name__)

        return "[yellow]" + "::".join(output) + "[/]"


@dtcl.dataclass(slots=True, repr=False, eq=False)
class _any_type_t(type_t):
    def Issues(self) -> list[str]:
        """"""
        return []

    def FirstAnnotationWithType(
        self, stripe: type[annotation_t], /
    ) -> annotation_t | None:
        """"""
        return None

    def ValueIsCompliant(self, value: Any, /) -> list[str]:
        """"""
        return []

    def InterpretedValueOf(self, value: Any, /) -> tuple[Any, list[str]]:
        """"""
        return value, []


ANY_TYPE = _any_type_t.NewFromTypeHint(Any)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from typing import Annotated, Sequence

from str_to_obj import annotation_t
from str_to_obj.type.hint import annotated_hint_t


@dtcl.dataclass(repr=False, eq=False)
class choices_t(annotation_t):
    ACCEPTED_TYPES = (str,)
    options: Sequence[str]

    @classmethod
    def NewAnnotatedType(cls, options: Sequence[str], /) -> annotated_hint_t:
        """"""
        return Annotated[str, cls(tuple(options))]

    def Issues(self) -> list[str]:
        """"""
        output = []

        for option in self.options:
            if not isinstance(option, str):
                output.append(
                    f"{type(option).__name__}: Invalid type of option {option} "
                    f"in {self}; Expected=str"
                )

        return output

    def ValueIsCompliant(self, value: str, /) -> list[str]:
        """"""
        if (self.options.__len__() == 0) or (value in self.options):
            # Options can be empty for a controlling parameter whose controlled section
            # has not been specified. In a GUI context, such controlled section should
            # have been populated programmatically.
            return []

        options = map(lambda _elm: f'"{_elm}"', self.options)
        options = " or ".join(options)

        return [f"{value}: Invalid choice; Expected={options}."]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import itertools as ittl
import textwrap as text
from typing import Any
from typing import NamedTuple as named_tuple_t
from typing import Sequence

from rich.text import Text as text_t
from str_to_obj.interface.console import TypeAsRichStr

from conf_ini_g.extension.string import AlignedOnSeparator
from conf_ini_g.phase.specification.base import base_t
from conf_ini_g.phase.specification.parameter.main import parameter_t


@dtcl.dataclass(init=False, repr=False, eq=False)  # Cannot have __init__ method
class controller_t(named_tuple_t):
    section: str = None
    parameter: str = None
    # All controller values are equal in role. However, this one is called primary
    # because it refers to the parameters of the section, as opposed to the parameters
    # in the alternatives.
    primary_value: Any = None

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        return (
            f"[magenta]{type(self).__name__}[/]: "
            f"[blue]{self.section}.{self.parameter}[/]={self.primary_value}"
        )


# TODO: Separate section into controlled and not controller, like the GUI classes.


@dtcl.dataclass(repr=False, eq=False)
class section_t(base_t, list[parameter_t]):
    """
    /!\ Must not be iterated classically, otherwise the alternatives, if any, will be
    missed.
    """

    # TODO: To avoid problems with the above remark, the class should be made uniterable.
    #     Then, the parameters attribute below could be used not only during instatiation,
    #     but constantly.
    category: str = "Main"
    optional: bool = False
    is_growable: bool = False  # True by default if no parameters are specified.
    controller: controller_t = None
    alternatives: dict[str, list[parameter_t]] = None
    parameters: list[parameter_t] | None = None  # Used only during instantiation.

    def __post_init__(self) -> None:
        """"""
        if self.parameters is None:
            self.is_growable = True
        else:
            self.extend(self.parameters)
            self.parameters = None

    @property
    def all_parameters(self) -> Sequence[parameter_t]:
        """"""
        if self.alternatives is None:
            return self
        else:
            return tuple(ittl.chain(self, *self.alternatives.values()))

    def ActiveParameters(self, controller_value: str, /) -> Sequence[parameter_t]:
        """"""
        if (self.alternatives is None) or (
            controller_value == self.controller.primary_value
        ):
            # Despite being a controlled section, there can be no alternatives if the
            # section was specified empty. In a GUI context, the section would have
            # been populated programmatically. In a CLI context, it is not.
            return self
        else:
            return self.alternatives[controller_value]

    @property
    def controlling_values(self) -> tuple[str, ...]:
        """
        No check of controlled status; Call with care.
        """
        if self.controller.primary_value is None:
            # The controlling values are not mentioned in the specification. They will
            # be set programmatically later on.
            return ()

        if self.alternatives is None:
            # The controlling values are not mentioned in the specification. They are
            # being set programmatically.
            return (self.controller.primary_value,)

        return (self.controller.primary_value,) + tuple(self.alternatives.keys())

    def Issues(self) -> list[str]:
        """"""
        output = super().Issues()

        valid_name_sets = [tuple(_prm.name for _prm in self)]
        if self.alternatives is not None:
            for parameters in self.alternatives.values():
                valid_name_sets.append(tuple(_prm.name for _prm in parameters))
        for valid_name_set in valid_name_sets:
            if valid_name_set.__len__() > set(valid_name_set).__len__():
                output.append(
                    f"{self.name}: Section with repeated parameter names (possibly in alternatives)"
                )

        basic = self.basic
        optional = self.optional

        if not (basic or optional):
            output.append(f"{self.name}: Section is not basic but not optional")

        if (self.controller is None) and (not optional) and (self.__len__() == 0):
            output.append(f"{self.name}: Empty mandatory section")

        n_parameters = 0
        n_basic_prms = 0
        for parameter in self.all_parameters:
            output.extend(f"/{self.name}/ {_iss}" for _iss in parameter.Issues())

            n_parameters += 1
            if parameter.basic:
                n_basic_prms += 1

            if parameter.basic and not basic:
                output.append(
                    f'{parameter.name}: Basic parameter in advanced section "{self.name}"'
                )
            if optional and not parameter.optional:
                output.append(
                    f'{parameter.name}: Mandatory parameter in optional section "{self.name}"'
                )

        if (n_parameters == 0) and not self.is_growable:
            output.append(
                f"{self.name}: Section without specified parameters which does not accept unspecified parameters"
            )
        if basic and (n_parameters > 0) and (n_basic_prms == 0):
            output.append(f"{self.name}: Basic section without any basic parameters")

        if self.controller is None:
            if self.alternatives is not None:
                output.append(f"{self.name}: Uncontrolled section with alternatives")
        else:
            if self.controller.section == self.name:
                output.append(f"{self.name}: Section cannot be controlled by itself")
            controlling_values = self.controlling_values
            if controlling_values.__len__() > set(controlling_values).__len__():
                output.append(
                    f"{self.name}: Controlled section has duplicated controller values"
                )
            for parameter in self.all_parameters:
                if not parameter.optional:
                    output.append(
                        f'{parameter.name}: Mandatory parameter in controlled section "{self.name}"'
                    )

            if self.controller.primary_value is None:
                if self.alternatives is not None:
                    output.append(
                        f"{self.name}: Controlled section without primary value"
                    )
                # else: This will be set programmatically later on.
            elif self.alternatives is None:
                output.append(f"{self.name}: Controlled section without alternatives")

        return output

    def _Item(self, key: str | int, /) -> parameter_t | None:
        """"""
        if isinstance(key, int):
            return list.__getitem__(self, key)

        for parameter in self.all_parameters:
            if parameter.name == key:
                return parameter

        return None

    def __contains__(self, key: str, /) -> bool:
        """"""
        return self._Item(key) is not None

    def __getitem__(self, key: str | int, /) -> parameter_t:
        """"""
        item = self._Item(key)
        if item is None:
            raise KeyError(f"{key}: Not a parameter of section {self.name}.")

        return item

    def AsDict(self) -> dict[str, Any]:
        """"""
        if self.controller is None:
            # TODO: Iteration over section will be replaced with something else when the class section will be modified.
            return {_prm.name: _prm.default for _prm in self if _prm.optional}
        else:
            # TODO: Iteration over section will be replaced with something else when the class section will be modified.
            output = {
                (self.controller.primary_value, _prm.name): _prm.default
                for _prm in self
                if _prm.optional
            }
            for value, parameters in self.alternatives.items():
                output.update(
                    {
                        (value, _prm.name): _prm.default
                        for _prm in parameters
                        if _prm.optional
                    }
                )

            return output

    def __rich__(self) -> str:
        """"""
        output = [
            TypeAsRichStr(self),
            *text.indent(super().__rich__(), "    ").splitlines(),
            f"    [blue]Category[/]@=@{self.category}",
            f"    [blue]Optional[/]@=@{self.optional}",
            f"    [blue]Growable[/]@=@{self.is_growable}",
        ]

        if has_controller := (self.controller is not None):
            output.append(f"    [blue]Controller[/]@=@{self.controller}")

        output.extend(text.indent(_prm.__rich__(), "    ") for _prm in self)
        if has_controller and (self.alternatives is not None):
            output.append("    With alternatives:")
            for ctl_name, parameters in self.alternatives.items():
                output.append(f"        {ctl_name}")
                for parameter in parameters:
                    output.append(text.indent(parameter.__rich__(), "            "))

        output = AlignedOnSeparator(output, "@=@", " = ")

        return "\n".join(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import textwrap as text
from typing import Sequence


def Flattened(string: str, /) -> str:
    """"""
    return text.dedent(string).replace("\n", "; ")


def AlignedOnSeparator(
    string: str | Sequence[str], separator: str, replacement: str, /
) -> str | tuple[str, ...] | list[str]:
    """"""
    if should_return_str := isinstance(string, str):
        lines = string.splitlines()
    else:
        lines = string
    indices = tuple(_lne.find(separator) for _lne in lines)
    longest = max(indices)

    output = (
        _lne.replace(separator, (longest - _lgt) * " " + replacement, 1)
        if _lgt > 0
        else _lne
        for _lne, _lgt in zip(lines, indices)
    )
    if should_return_str:
        return "\n".join(output)
    elif isinstance(string, tuple):
        return tuple(output)
    else:
        return list(output)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
import inspect as nspt
from pathlib import Path as pl_path_t
from typing import Any, Callable, Iterator, Sequence

from babelwidget.backend.generic.path_chooser import NewSelectedOutputDocument
from babelwidget.main import backend_t
from babelwidget.main import base_h as library_wgt_h

from conf_ini_g.interface.storage.config import INIConfig, SaveConfigToINIDocument
from conf_ini_g.interface.window.component.config.action import ActionButtonsLayout
from conf_ini_g.interface.window.component.config.advanced import AdvancedModeLayout
from conf_ini_g.interface.window.component.config.title import TitleLayout
from conf_ini_g.interface.window.component.section.collection import (
    SectionsAndCategories,
    any_section_h,
)
from conf_ini_g.interface.window.component.section.main import controlled_section_t
from conf_ini_g.phase.generic.config import AsStr
from conf_ini_g.phase.specification.config.main import config_t as config_spec_t
from conf_ini_g.phase.specification.parameter.main import (
    parameter_t as parameter_spec_t,
)
from conf_ini_g.phase.specification.section.main import section_t as section_spec_t
from conf_ini_g.phase.typed.config.main import (
    TypedFromUntyped,
    WithConsumedUnits,
    typed_config_h,
)
from conf_ini_g.phase.untyped.config import ini_config_h


@dtcl.dataclass(repr=False, eq=False)
class config_t:
    """
    The class cannot use slots because it disables weak referencing, which is required.
    See error message below when using slots:
    TypeError: cannot create weak reference to 'config_t' object
    [...]
    File "[...]conf_ini_g/catalog/interface/window/backend/pyqt5/widget/choices.py", line 41, in SetFunction
        self.clicked.connect(function)
        │                    └ <bound method config_t.ToggleAdvancedMode of <conf_ini_g.interface.window.config.config_t object at [...]>>
        └ <conf_ini_g.catalog.interface.window.backend.pyqt5.widget.choices.radio_choice_wgt_h object at [...]>

    Widget might not cooperate well with list, in which case Python raises the
    following exception: TypeError: multiple bases have instance lay-out conflict
    To be safe, "sections" is a field instead of being part of the class definition.
    """

    config_spec: config_spec_t
    ini_path: pl_path_t | None
    backend: backend_t
    library_wgt: library_wgt_h
    UpdateHistory: Callable[[pl_path_t | str], None] | None
    #
    Action: Callable[[typed_config_h], None] = None
    #
    sections: dict[str, any_section_h] = dtcl.field(init=False, default=None)
    # Both an access for interacting with widgets, and a reference keeper to prevent
    # autonomous widgets from loosing their "liveness".
    _widget: dict[str, library_wgt_h] = dtcl.field(init=False, default_factory=dict)

    @classmethod
    def NewForSpecification(
        cls,
        title: str | None,
        config_spec: config_spec_t,
        backend: backend_t,
        ini_path: pl_path_t | None,
        /,
        *,
        history: Sequence[str] | None = None,
        UpdateHistory: Callable[[pl_path_t | str], None] | None = None,
        advanced_mode: bool = False,
        action: tuple[str, Callable[[typed_config_h], None]] = None,
    ) -> config_t:
        """"""
        if ini_path is not None:
            if UpdateHistory is not None:
                UpdateHistory(ini_path)
            as_str = str(ini_path)
            if history is None:
                history = (as_str,)
            elif as_str not in history:
                history = list(history)
                history.append(as_str)

        if action is None:
            kwargs = {}
        else:
            kwargs = {"Action": action[1]}

        output = cls(
            config_spec=config_spec,
            ini_path=ini_path,
            backend=backend,
            library_wgt=backend.base_t(),
            UpdateHistory=UpdateHistory,
            **kwargs,
        )

        # --- Top-level widgets
        (
            title_lyt,
            ini_path_wgt,
            history_button,
            history_menu,
            close_button,
        ) = TitleLayout(
            title,
            config_spec,
            history,
            backend,
            ini_path,
            output.UpdateWithNewINI,
            output.library_wgt.close,
        )
        advanced_mode_lyt, adv_mode_wgt = AdvancedModeLayout(
            advanced_mode, backend, output.ToggleAdvancedMode
        )
        button_lyt, action_button, action_wgts = ActionButtonsLayout(
            action,
            ini_path is not None,
            backend,
            output.ShowInINIFormat,
            output.SaveConfig,
            output.LaunchAction,
            output.library_wgt.close,
        )
        output._widget["ini_path"] = ini_path_wgt
        output._widget["history_button"] = history_button
        output._widget["history_menu"] = history_menu
        output._widget["adv_mode"] = adv_mode_wgt
        output._widget["action"] = action_button
        output._widget["action_buttons"] = action_wgts
        output._widget["close"] = close_button

        # --- Sections
        output.sections, controlled_sections, category_selector = SectionsAndCategories(
            config_spec, None, backend
        )
        output._AssociateSectionsAndControllers(controlled_sections)
        output._widget["category_selector"] = category_selector

        # --- Layout...
        layout = backend.grid_lyt_t()
        if title_lyt is None:
            first_available_row = 0
        else:
            layout.addLayout(title_lyt, 0, 0, 1, 1)
            first_available_row = 1
        layout.addWidget(category_selector, first_available_row, 0, 1, 1)
        layout.addLayout(advanced_mode_lyt, first_available_row + 1, 0, 1, 1)
        layout.addLayout(button_lyt, first_available_row + 2, 0, 1, 1)

        output.library_wgt.setLayout(layout)
        # --- ...Layout

        # output.ToggleAdvancedMode(advanced_mode)

        return output

    def _AssociateSectionsAndControllers(
        self,
        controlled_sections: Sequence[tuple[controlled_section_t, section_spec_t]],
        /,
    ) -> None:
        """"""
        for section, section_spec in controlled_sections:
            controller = section_spec.controller
            # TODO: Check that controller section cannot be controlled. Is it done already?
            parameter = self[controller.section].active_parameters[controller.parameter]
            value_wgt = parameter.value
            if hasattr(value_wgt, "SetFunction"):
                value_wgt.SetFunction(section.page_stack.setCurrentIndex)
            else:
                self.backend.ShowErrorMessage(
                    f"{controller.section}.{controller.parameter}: "
                    f'Controller has no "SetFunction" method; Disabling control.'
                )

    def ReassignCloseButtonTarget(self) -> None:
        """"""
        current = self.library_wgt
        while current is not None:
            main_window = current
            current = current.parent()

        self._widget["close"].SetFunction(main_window.close)

    def SetValues(
        self, config: typed_config_h, units: dict[tuple[str, str], tuple[str, float]], /
    ) -> None:
        """"""
        for sct_name, section in self.sections.items():
            for parameter_set in section.all_parameters:
                for prm_name, parameter in parameter_set.items():
                    parameter_spec = self.config_spec[sct_name][prm_name]
                    typed_section = config[sct_name]
                    if prm_name in typed_section:
                        parameter.value.Assign(
                            typed_section[prm_name], parameter_spec.type
                        )
                        unit = units.get((sct_name, prm_name), None)
                        if unit is not None:
                            parameter.unit.setText(unit[0])
                    elif parameter_spec.optional:
                        parameter.value.Assign(
                            parameter_spec.default, parameter_spec.type
                        )

    def ToggleAdvancedMode(self, advanced_mode: bool, /) -> None:
        """"""
        for section_name, section in self.sections.items():
            section_spec = self.config_spec[section_name]
            if section_spec.basic:
                should_check_parameters = True
            elif advanced_mode:
                section.library_wgt.setVisible(True)
                should_check_parameters = True
            else:
                section.library_wgt.setVisible(False)
                should_check_parameters = False

            if should_check_parameters:
                parameters = section.active_parameters.values()
                parameter_specs = self._ActiveParameterSpcOfSection(section_spec)
                for parameter, parameter_spec in zip(parameters, parameter_specs):
                    if not parameter_spec.basic:
                        if advanced_mode:
                            parameter.SetVisible(True)
                        else:
                            parameter.SetVisible(False)

    def UpdateWithNewINI(self, ini_path: pl_path_t | str, /) -> None:
        """"""
        if isinstance(ini_path, str):
            ini_path = pl_path_t(ini_path)

        config_ini = INIConfig(ini_path)
        config_typed, config_units, config_comments, issues = TypedFromUntyped(
            config_ini, None, self.config_spec
        )
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        category_selector = self._widget["category_selector"]
        if isinstance(category_selector, self.backend.tabs_t):
            # Note: idx^th layout: category_selector.widget(t_idx).widget().layout().
            while category_selector.count() > 0:
                category_selector.removeTab(0)
        else:
            layout = category_selector.widget().layout()
            while layout.count() > 0:
                layout.itemAt(0).widget().setParent(None)

        self.sections, controlled_sections, _ = SectionsAndCategories(
            self.config_spec, category_selector, self.backend
        )
        self._AssociateSectionsAndControllers(controlled_sections)
        self.SetValues(config_typed, config_units)
        self.ToggleAdvancedMode(self._widget["adv_mode"].true_btn.isChecked())

        self._widget["ini_path"].Assign(ini_path, None)
        self._widget["history_button"].setEnabled(True)
        if str(ini_path) not in (
            _elm.text() for _elm in self._widget["history_menu"].actions()
        ):
            self._widget["history_menu"].addAction(str(ini_path))

        if self.UpdateHistory is not None:
            self.UpdateHistory(ini_path)

    def _ActiveParameterSpcOfSection(
        self, section_spec: section_spec_t, /
    ) -> Sequence[parameter_spec_t]:
        """"""
        if (controller := section_spec.controller) is None:
            output = section_spec
        else:
            controller = self[controller.section].active_parameters[
                controller.parameter
            ]
            output = section_spec.ActiveParameters(controller.Text())

        return output

    def AsINI(self) -> ini_config_h:
        """"""
        output = {}

        for section_name, section in self.sections.items():
            section_spec = self.config_spec[section_name]
            parameter_specs = self._ActiveParameterSpcOfSection(section_spec)

            output[section_name] = section.AsINI(parameter_specs)

        return output

    def LaunchAction(self) -> None:
        """"""
        config_full, units_full, _, issues = TypedFromUntyped(
            self.AsINI(), None, self.config_spec
        )
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        typed_config, issues = WithConsumedUnits(config_full, units_full)
        if issues.__len__() > 0:
            self.backend.ShowErrorMessage("\n".join(issues), parent=self.library_wgt)
            return

        self.setEnabled(False)
        self.backend.qt_core_app_t.processEvents()
        try:
            self.Action(typed_config)
        except Exception as exception:
            trace = nspt.trace()[-1]
            context = "\n".join(trace.code_context)
            self.backend.ShowErrorMessage(
                f"{trace.filename}@{trace.lineno}:{trace.function}\n"
                f"{context}\n"
                f"{exception}",
                parent=self.library_wgt,
            )
        self.setEnabled(True)

    def ShowInINIFormat(self) -> None:
        """"""
        config = AsStr(self.AsINI(), color="html")
        self.backend.ShowMessage("INI Config", "<tt>" + config + "<tt/>")

    def SaveConfig(self, new_ini: bool, /) -> None:
        """"""
        if new_ini:
            path = NewSelectedOutputDocument(
                "Save Config As",
                "Save Config As",
                self.backend,
                mode="document",
                valid_types={"Config files": ("ini", "INI")},
            )
        else:
            path = self.ini_path  # Will overwrite current INI document

        if path is not None:
            config = self.AsINI()
            issues = SaveConfigToINIDocument(config, path)
            if issues.__len__() > 0:
                self.backend.ShowErrorMessage("\n".join(issues), parent=self)
            else:
                self.ini_path = path

    def __getattr__(self, attribute: str, /) -> Any:
        """
        E.g., used for "show".
        """
        try:
            output = super().__getattribute__(attribute)
        except AttributeError:
            output = getattr(self.library_wgt, attribute)

        return output

    def __contains__(self, key: str, /) -> bool:
        """"""
        return key in self.sections

    def __getitem__(self, key: str, /) -> any_section_h:
        """"""
        return self.sections[key]

    def __iter__(self) -> Iterator[str]:
        """"""
        return self.sections.keys()



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
from enum import Enum as enum_t
from typing import Annotated, ClassVar

from str_to_obj import annotation_t
from str_to_obj.type.hint import annotated_hint_t


@dtcl.dataclass(slots=True, repr=False, eq=False)
class boolean_t(annotation_t):
    class MODE(enum_t):
        # Always list true value first
        true_false = ("True", "False")
        yes_no = ("Yes", "No")
        on_off = ("On", "Off")

    _MODES: ClassVar[tuple[str, ...]] = tuple(_elm for _elm in MODE.__members__.keys())

    # TODO: Why is ACCEPTED_TYPES not used? In Issues? In valueIsCompliant? Check
    #     similar TODO in str-to-obj.
    ACCEPTED_TYPES = (bool,)
    mode: enum_t = MODE.true_false

    @classmethod
    def NewAnnotatedType(
        cls, /, *, mode: enum_t | str | None = None
    ) -> annotated_hint_t:
        """"""
        if mode is None:
            mode = cls.MODE.true_false
        elif isinstance(mode, str):
            if mode in cls._MODES:
                mode = cls.MODE[mode].value
            else:
                valid = " or ".join(cls._MODES)
                raise ValueError(
                    f"Invalid boolean mode: Actual={mode}; Expected={valid}."
                )

        return Annotated[bool, cls(mode=mode)]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dataclasses as dtcl
import string as strg

from rich.text import Text as text_t

# Must not be <space> to allow assignment through command-line arguments.
_WORD_SEPARATOR = "_"
_VALID_CHARACTERS = strg.ascii_letters + strg.digits + _WORD_SEPARATOR


@dtcl.dataclass(repr=False, eq=False)
class base_t:
    name: str
    definition: str = "No Definition Provided"
    description: str = "No Description Provided"
    basic: bool = True

    def Issues(self) -> list[str]:
        """"""
        output = []

        if any(_chr not in _VALID_CHARACTERS for _chr in self.name):
            output.append(
                f"{self.name}: Name contains invalid characters; "
                f"Valid={_VALID_CHARACTERS}"
            )
        if not isinstance(self.definition, str):
            output.append(
                f'{self.definition}: Not a valid string for definition of "{self.name}"'
            )
        if not isinstance(self.description, str):
            output.append(
                f'{self.description}: Not a valid string for description of "{self.name}"'
            )

        return output

    def __str__(self) -> str:
        """"""
        return text_t.from_markup(self.__rich__()).plain

    def __rich__(self) -> str:
        """"""
        return f"[blue]Name[/]@=@{self.name}\n[blue]Basic[/]@=@{self.basic}"
